"""Module to perform annual chain linking with PySpark.

This module provides the functions:

* :func:`chain_linking_window`
* :func:`chain_linking_join`

Chain linking is the process of joining together two indices that
overlap in one period by rescaling one of them to make its value equal
to that of the other in the same period thus combining them into a
single time series. This is achieved by multiplying the index value by
the linking factor.
"""
from pyspark.sql import (
    DataFrame as SparkDF,
    functions as F,
    Window,
)

from .helpers import _cumprod_over_period

__author__ = ['Dr. Usman Kayani']

def chain_linking_window(
    df: SparkDF,
    date_col: str = 'period',
) -> SparkDF:
    """Chain linking method using only window functions.

    This method uses a cumulative product of unchained January indices
    to determine the linking factors, from a window partitioned by year,
    and a lag function over a period window to calculate the chained
    indices using pointwise multiplication of columns (Hadarmard
    product).

    Parameters
    ----------
    df: SparkDF
        Dataframe containing the timeseries and unchained index values.
    date_col: str, default 'period'
        User-defined name for the date column.

    Returns
    -------
    SparkDF
        Dataframe containing the timeseries and chained index values.
    """
    # Determine linking factors from cumulative product of unchained
    # January indices, which we determine by a window function
    # partitioned by year and by selecting the first value.
    window_year = Window.partitionBy(F.year(date_col)).orderBy(date_col)
    link_index = F.first('index_value').over(window_year)
    linking_factor = _cumprod_over_period(link_index/100)

    # Lag linking factors by one month to adjust for the correct
    # periods.
    lagged_linking_factor = (
        F.lag(linking_factor, 1, 1).over(Window.orderBy(date_col))
    )

    # Determine chained indices from unchained indices multiplied by a
    # link factor.
    chained_indices = F.col('index_value') * lagged_linking_factor

    return df.withColumn('index_value', chained_indices)


def chain_linking_join(
    df: SparkDF,
    date_col: str = 'period',
    link_month: int = 1,
) -> SparkDF:
    """
    Chain linking method using lag over window and join in PySpark.

    This method uses a cumulative product of unchained January indices
    to determine the linking factors and a lag function over a period
    window to calculate the chained indices using pointwise
    multiplication of columns (Hadarmard product).

    Parameters
    ----------
    df: SparkDF
        Dataframe containing the timeseries and unchained index values.
    date_col: str, default 'period'
        User-defined name for the date column.
    link_month: int, default 1
        Linking month for chaining.

    Returns
    -------
    SparkDF
        Dataframe containing the timeseries and chained index values.
    """
    # Determine linking factors from cumulative product of unchained January
    # indices.
    df_link_month = (
            df
            .filter(F.month(date_col) == link_month)
            .withColumn('year', F.year(date_col))
            .withColumn(
                'link_factor',
                _cumprod_over_period(F.col('index_value') / 100, date_col)
            )
            .drop(date_col, 'index_value')
    )

    # Determine chained indices from unchained multiplied by a link factor.
    return (
        df
        .withColumn('year', F.year(date_col))
        .join(df_link_month, on='year')
        .withColumn(
            'link_factor',
            F.lag('link_factor', 1, 1).over(Window.orderBy(date_col))
        )
        .withColumn(
            'index_value',
            F.col('index_value') * F.col('link_factor')
        )
        .drop('link_factor', 'year')
    )





from pyspark.sql import (
    functions as F,
    Column as SparkCol,
)
from pyspark.sql.window import Window

__author__ = ['Dr. Usman Kayani']

def _weights_calc_pyspark(
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
)-> SparkCol:
    """Calculate weights from expenditure shares in PySpark."""
    window = Window.partitionBy(date_col)
    expenditure = F.col(price_col)*F.col(quantity_col)
    return expenditure / F.sum(expenditure).over(window)
  
def _cumprod_over_period(
    col: SparkCol,
    date_col: str = 'period'
) -> SparkCol:
    """Cumulative product of numeric column over a period window."""
    window = (
        Window
        .orderBy(date_col)
        .rowsBetween(Window.unboundedPreceding, Window.currentRow)
    )
    return F.exp(F.sum(F.log(col)).over(window))



"""Module to perform annual chain linking with Pandas.

This module provides the functions:

* :func:`chain_linking_pandas_join`
* :func:`chain_linking_pandas_matrix`

Chain linking is the process of joining together two indices that
overlap in one period by rescaling one of them to make its value equal
to that of the other in the same period thus combining them into a
single time series. This is achieved by multiplying the index value by
the linking factor.
"""
from typing import List
import pandas as pd
import numpy as np

__author__ = ['Dr. Usman Kayani']

def chain_linking_pandas_join(
    dfs: List[pd.DataFrame],
    date_col: str = 'period',
    link_month: int = 1,
) -> pd.DataFrame:
    """
    Chain linking method using pandas shift and join in pandas.

    This method uses a cumulative product of unchained January indices
    to determine the linking factors and a shift function to calculate
    the chained indices.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing the timeseries and unchained index values.
    date_col: str, default 'period'
        User-defined name for the date column.
    link_month: int, default 1
        Linking month for chaining.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the timeseries and chained index values.
    """
    df, df_link_month = dfs
    # Determine chained indices from unchained multiplied by a link factor.
    df = (
        df
        .assign(year=df[date_col].dt.year)
        .merge(df_link_month, on='year')
    )
    df['link_factor'] = df.link_factor.shift(1, fill_value=1)
    df['index_value'] = df.index_value * df.link_factor

    return df.drop(columns=['link_factor', 'year'])


def chain_linking_pandas_matrix(
    dfs: List[pd.DataFrame],
    date_col: str = 'period',
    link_month: int = 1,
) -> pd.DataFrame:
    """
    Chain linking method using matrix operations with pandas and numpy.

    This method uses a cumulative product of unchained January indices
    to determine the linking factors, then converting these to an array
    with repeated values to calculate the chained indices via pointwise
    multiplication (Hadamard product).

    Parameters
    ----------
    df: Pandas DataFrame
        Dataframe containing the timeseries and unchained index values.
    date_col: str, default 'period'
        User-defined name for the date column.
    link_month: int, default 1
        Linking month for chaining.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the timeseries and chained index values.
    """
    df, df_link_month = dfs
    # Convert link factors to an array for pointwise multiplication.
    if len(df) % 12 == 1:
        link_factor = df_link_month.link_factor.to_numpy()[:-1]
    else:
        link_factor = df_link_month.link_factor.to_numpy()

    # Repeat link factors for 12 months (Feb(year)-Jan(year+1)) and
    # insert 1 for the first.
    link_factor = np.insert(np.repeat(link_factor, 12), 0, 1)

    # Determine chained indices from unchained multiplied by a link factor.
    return df.assign(index_value=df.index_value * link_factor)

def linking_method(
    df: pd.DataFrame,
    date_col: str = 'period',
    link_month: int = 1,
    method: str = 'pandas_join',
) -> pd.DataFrame:
    """
    Chain linking method using pandas shift and join in pandas.

    This method uses a cumulative product of unchained January indices
    to determine the linking factors and a shift function to calculate
    the chained indices.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing the timeseries and unchained index values.
    date_col: str, default 'period'
        User-defined name for the date column.
    link_month: int, default 1
        Linking month for chaining.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the timeseries and chained index values.
    """
    # Conver date column to datetime for computation.
    df[date_col] = pd.to_datetime(df[date_col])

    # Determine linking factors from cumulative product of unchained January
    # indices.
    df_link_month = df.loc[df[date_col].dt.month == link_month]
    df_link_month = (
        df_link_month
        .assign(link_factor=(df_link_month.index_value / 100).cumprod())
        .drop(columns=[date_col, 'index_value'])
    )
    # Determine the linking method.
    if method == 'pandas_join':
        return chain_linking_pandas_join(
            dfs=[df, df_link_month],
            date_col=date_col,
            link_month=link_month,
        )
    elif method == 'pandas_matrix':
        return chain_linking_pandas_matrix(
            dfs=[df, df_link_month],
            date_col=date_col,
            link_month=link_month,
        )
    else:
        raise ValueError(
            'Invalid linking method. '
            'Valid methods are "pandas_join" and "pandas_matrix".'
        )


"""
Provides the following bilateral methods:

* :func:`carli`
* :func:`jevons`
* :func:`dutot`
* :func:`lowe`
* :func:`laspeyres`
* :func:`paasche`
* :func:`geom_laspeyres`
* :func:`geom_paasche`
* :func:`drobish`
* :func:`marshall_edgeworth`
* :func:`palgrave`
* :func:`fisher`
* :func:`tornqvist`
* :func:`walsh`
* :func:`sato_vartia`
* :func:`geary_khamis_b`
* :func:`tpd`
* :func:`rothwell`
* :func:`young`
"""
import numpy as np

__author__ = ['Dr. Usman Kayani']

def carli(
    p0: np.array, 
    p1: np.array,
) -> float:
    """
    Carli bilateral index, using price information.
    
    .. math::
        \\text{Carli} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    """
    return np.mean(p1 / p0)

def jevons(
    p0: np.array, 
    p1: np.array
) -> float:
    """
    Jevons bilateral index, using price information.
    
    .. math::
        \\text{Jevons} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    """
    return np.prod((p1 / p0) ** (1 / len(p0)))

def dutot(
    p0: np.array, 
    p1: np.array
) -> float:
    """
    Dutot bilateral index, using price information.
    
    .. math::
        \\text{Dutot} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    """
    return np.sum(p1)/np.sum(p0)

def laspeyres(
    p0: np.array, 
    p1: np.array,
    q0: np.array,
) -> float:
    """
    Laspeyres bilateral index, using price and base quantity information.
    
    .. math::
        \\text{Laspeyres} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    """
    return lowe(p0, p1, q0)

def paasche(
    p0: np.array, 
    p1: np.array,
    q1: np.array,
) -> float:
    """Paasche bilateral index, using price and current quantity information."""
    return lowe(p0, p1, q1)

def geom_laspeyres(
    p0: np.array,
    p1: np.array,
    q0: np.array,
) -> float:
    """
    Geometric Laspeyres bilateral index, using price and base quantity information.
    
    .. math::
        \\text{Geometric Laspeyres} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    """
    s0 = (p0 * q0) / np.sum(p0 * q0)
    return np.prod((p1 / p0) ** s0)

def geom_paasche(
    p0: np.array, 
    p1: np.array,
    q1: np.array,
) -> float:
    """
    Geometric Paasche bilateral index, using price and base quantity information.

    .. math::
        \\text{Geometric Paasche} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q1: Base quantity vector.
    """
    s1 = (p1 * q1) / np.sum(p1 * q1)
    return np.prod((p1 / p0) ** s1)

def lowe(
    p0: np.array, 
    p1: np.array,
    q: np.array,
) -> float:
    """
    Lowe bilateral index, using price and arbitrary quantity information.
    
    .. math::
        \\text{Lowe} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q: Arbitrary quantity vector.
    """
    return np.sum(p1 * q) / np.sum(p0 * q)

def young(
    p0: np.array,
    p1: np.array,
    p: np.array, 
    q: np.array,
) -> float:
    """
    Young bilateral index, using price and arbitrary price and quantity information.
    
    .. math::

        \\text{Young} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param p: Arbitrary price vector.
    :param q: Arbitrary quantity vector.
    """
    s = (p * q) / np.sum(p * q)
    return np.sum(s * (p1 / p0))

def drobish(
    p0: np.array, 
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Drobish bilateral index, using price and quantity information.
    
    .. math::
        \\text{Drobish} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    return (laspeyres(p0, p1, q0) + paasche(p0, p1, q1)) / 2

def marshall_edgeworth(
    p0: np.array,
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Marshall-Edgeworth bilateral index, using price and quantity information.
    
    .. math::
        \\text{Marshall-Edgeworth} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    return np.sum(p1 * (q0 + q1)) / np.sum(p0 * (q0 + q1))

def palgrave(
    p0: np.array,
    p1: np.array,
    q1: np.array,
) -> float:
    """
    Palgrave bilateral index, using price and arbitrary price and quantity information.
    
    .. math::
        \\text{Palgrave} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q1: Arbitrary quantity vector.
    """
    s1 = (p1 * q1) / np.sum(p1 * q1)
    return np.sum(s1 * (p1 / p0))

def fisher(
    p0: np.array, 
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Fisher bilateral index, using price and quantity information.
    
    .. math::
        \\text{Fisher} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    """
    return np.sqrt(laspeyres(p0, p1, q0) * paasche(p0, p1, q1))

def tornqvist(
    p0: np.array,
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Torqvist bilateral index, using price and quantity information.
    
    .. math::
        \\text{Torqvist} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    s0 = (p0 * q0) / np.sum(p0 * q0)
    s1 = (p1 * q1) / np.sum(p1 * q1)
    return np.prod((p1 / p0) ** (0.5 * (s0 + s1)))

def walsh(
    p0: np.array, 
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Walsh bilateral index, using price and quantity information.
    
    .. math::
        \\text{Walsh} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    return np.sum(np.sqrt(q0 * q1).dot(p1)) / np.sum(np.sqrt(q0 * q1).dot(p0))

def sato_vartia(
    p0: np.array,
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Sato-Vartia bilateral index, using price and quantity information.
    
    .. math::
        \\text{Sato-Vartia} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    s0 = (p0 * q0) / np.sum(p0 * q0)
    s1 = (p1 * q1) / np.sum(p1 * q1)
    w = (s1 - s0) / (np.log(s1) - np.log(s0))
    w = w / np.sum(w)
    return np.prod((p1 / p0) ** w)

def geary_khamis_b(
    p0: np.array,
    p1: np.array,
    q0: np.array,
    q1: np.array,
) -> float:
    """
    Geary-Khamis bilateral index, using price and quantity information.
    
    .. math::
        \\text{Geary-Khamis} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    :param q1: Current quantity vector.
    """
    hmean = (2 * q0 * q1) / (q0 + q1)
    return np.sum(hmean * p1) / np.sum(hmean * p0)

def rothwell(
    p0: np.array,
    p1: np.array,
    q0: np.array,
) -> float:
    """
    Rothwell bilateral index, using price and base quantity information.
    
    .. math::
        \\text{Rothwell} = \\frac{\\sum_{i=1}^{n} p_i}{\\sum_{i=1}^{n} p_0}

    :param p0: Base price vector.
    :param p1: Current price vector.
    :param q0: Base quantity vector.
    """
    pu = np.sum(p0 * q0) / np.sum(q0)
    return np.sum(p1 * q0) / np.sum(pu * q0)



from typing import Sequence, Tuple

import pandas as pd
import numpy as np
import statsmodels.api as SM
from sklearn.linear_model import LinearRegression as LR

from .helpers import _vars_split

__author__ = ['Dr. Usman Kayani']

def wls(
    df: pd.DataFrame,
    dependent_var: str,
    independent_vars: Sequence[str],
    engine: str = 'numpy',
) -> pd.Series:
    """Perform weighted least squares regression.

    Parameters
    ----------
    df : pd.DataFrame
        Contains columns for each object in the formula and a weights
        column as a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.
    engine: str, defaults to 'numpy'
        Options: {'numpy', 'statsmodels', 'sklearn'}

        Engine to use for wls computation.

    Returns
    -------
    pd.Series
        Coefficients for a weighted linear regression model derived from a
        least-squares fit.

    """
    return globals()[f'wls_{engine}'](df, dependent_var, independent_vars)

def wls_numpy(
    df: pd.DataFrame,
    dependent_var: str,
    independent_vars: Sequence[str],
) -> pd.Series:
    """
    Weighted least squares with numpy.
    
    Parameters
    ----------
    df : pd.DataFrame
        Contains columns for each object in the formula and a weights
        column as a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.

    Returns
    -------
    pd.Series
        Coefficients for a weighted linear regression model derived from a
        least-squares fit.
    """
    # Obtain variables and weight for wls regression.
    X, Y, weights = _get_vars(df, dependent_var, independent_vars)
    W = np.array(weights)

    # Matrix WLS to fit model.
    coeffs = np.linalg.pinv(X.T @ (W[:, None] * X)) @ (X.T @ (W * Y))    
    
    return pd.Series(coeffs, index=X.columns)

def wls_statsmodels(
    df: pd.DataFrame,
    dependent_var: str,
    independent_vars: Sequence[str],
) -> pd.Series:
    """
    Weighted least squares with statsmodels.
    
    Parameters
    ----------
    df : pd.DataFrame
        Contains columns for each object in the formula and a weights
        column as a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.
    
    Returns
    -------
    pd.Series
        Coefficients for a weighted linear regression model derived from a
        least-squares fit.
    """
    # Obtain variables and weight for wls regression.
    X, Y, weights = _get_vars(df, dependent_var, independent_vars)
    
    # Statsmodels WLS to fit model.
    model = SM.WLS(Y,X, weights=weights).fit(params_only=True)
    return model.params

def wls_sklearn(
    df: pd.DataFrame,
    dependent_var: str,
    independent_vars: Sequence[str],
) -> pd.Series:
    """
    Weighted least squares with Sklearn.
    
    Parameters
    ----------
    df : pd.DataFrame
        Contains columns for each object in the formula and a weights
        column as a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.
    
    Returns
    -------
    pd.Series
        Coefficients for a weighted linear regression model derived from a
        least-squares fit.
    """
    # Obtain variables and weight for wls regression.
    X, Y, weights = _get_vars(df, dependent_var, independent_vars)

    # Sklearn WLS to fit model.
    coefficients = (
        LR(fit_intercept=False)
        .fit(X, Y, sample_weight=weights)
        .coef_
    )

    return pd.Series(coefficients, index=X.columns)

def _get_vars(
    df: pd.DataFrame,
    dependent_var: str,
    independent_vars: Sequence[str],
) -> Tuple[pd.Series, pd.Series, pd.Series]:
    """
    Get variables for least squares regression.
    
    Parameters
    ----------
    df : pd.DataFrame
        Contains columns for each object in the formula and a weights
        column as a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.

    Returns
    -------
    tuple of pd.Series
        X, Y, and weights for least squares regression.

    """
    # Split the categoerical and numerical vars.
    categorical_vars, numerical_vars = _vars_split(df[independent_vars[1:]])

    # Convert categorical variables to dummy.
    df_dummy = pd.get_dummies(
        df,
        columns=[independent_vars[0], *categorical_vars],
        drop_first=True
    )

    # Create a column with all values equal to 1 for the constant term in the regression.
    df_dummy['const'] = 1

    # Determine the independent variables.
    vars = df_dummy.columns.difference(df.columns).tolist() + numerical_vars

    return df_dummy[vars], df_dummy[dependent_var], df_dummy['weights']



"""Geary-Khamism (pure) index in PySpark using a native PySpark method.

Provides the following function:

* :func:`geary_khamis`
"""
from typing import (
    Optional,
    Sequence,
    Union,
)

from pyspark.sql import (
    Column as SparkCol,
    DataFrame as SparkDF,
    functions as F,
    Window,
    WindowSpec,
    SparkSession
)

import pandas as pd

def get_window_spec(groups: Optional[Sequence[str]] = None) -> WindowSpec:
    """Return WindowSpec partitioned by levels, defaulting to whole df."""
    return Window.partitionBy(groups) if groups else Window.partitionBy()

def get_weight_shares(
    weights: str,
    levels: Optional[Union[str, Sequence[str]]] = None,
) -> SparkCol:
    """Divide weights by sum of weights for each group."""
    return weights / F.sum(weights).over(get_window_spec(levels))

def geary_khamis_pure(
    df: SparkDF,
    groups: Union[str, Sequence[str]] = None,
    no_of_iterations: int = 5,
    precision: float = None,
    checkpoint: bool = False,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
    product_id_col: str = 'id',
) -> SparkDF:
    """
    Calculate the index values with the Geary-Khamis iterative method using
    native Spark functionality.
    """
    index_groups = [*groups, date_col] if groups else [date_col]
    levels = [*groups, product_id_col] if groups else [product_id_col]

    # Define and price and quantities vectors.
    prices = F.col(price_col)
    quantities = F.col(quantity_col)

    # Initialise price levels as 1's to find the solution with
    # iteration.
    price_levels = F.lit(1)
    if precision:
        df = df.withColumn('price_levels', price_levels)

    # Define windows for sum over groups.
    window_levels = get_window_spec(levels)
    window_id = get_window_spec([product_id_col])
    window_index_groups = get_window_spec(index_groups)
    window_groups = get_window_spec(groups).orderBy(date_col)

    # Calculate the quantity share and turnover for the iteration.
    quantity_share = quantities / F.sum(quantities).over(window_levels)
    turnover = F.sum(prices*quantities).over(window_index_groups)

    # Iterate until we reach the set level of precision, or after a set
    # number of iterations if they do not converge.
    for iteration in range(no_of_iterations):
        # Obtain matrices for iterative calculation.
        deflated_prices = prices / price_levels
        factors = F.sum(deflated_prices * quantity_share).over(window_id)

        # Determine the weighted quantity index for current price
        # levels.
        weighted_quality_index = (
            F.sum(quantities * factors)
            .over(window_index_groups)
        )

        # Calculate new price levels from previous value.
        new_price_levels = turnover / weighted_quality_index
        df = df.withColumn('new_price_levels', new_price_levels)

        if precision:
            pl_diff = F.col('new_price_levels') - F.col('price_levels')
            is_precise = (
                not bool(df.filter(F.abs(pl_diff) > precision).take(1))
            )
            if is_precise:
                print(f'Precision of {precision} in {iteration} iterations.')
                break

        # Otherwise set price level for next iteration. We set this from
        # the column values appended to the df above, rather than the
        # SparkCol object as this works much faster within Spark.
        price_levels = F.col('new_price_levels')
        df = df.withColumn('price_levels', price_levels)

        # todo: add a check for precision?
        if checkpoint and iteration % 10 == 0:
            df = df.checkpoint()

    # Normalize by first period for final output.
    indices = price_levels / F.first(price_levels).over(window_groups)

    return (
        df
        .withColumn('index_value', indices)
        .groupby(index_groups)
        .agg(F.first('index_value').alias('index_value'))
        .orderBy(index_groups)
    )

df = pd.read_csv('tests/test_data/large_input_df.csv')

spark = SparkSession.builder.master('local').getOrCreate()

df = spark.createDataFrame(df)

geary_khamis_pure(df, groups=['group']).show(100)


"""
Multilateral methods using native PySpark functionality. 

Provides the following function:

* :func:`multilateral_methods_pyspark`
"""
from typing import Sequence, Optional

import pandas as pd
from pyspark.sql import (
    DataFrame as SparkDF,
)

from .helpers import _weights_calc_pyspark
from .multilateral import geary_khamis_pyspark, time_dummy_pyspark

__author__ = ['Dr. Usman Kayani']

def multilateral_methods_pyspark(
    df: SparkDF,
    method: str,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
    product_id_col: str = 'id',
    characteristics: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    """
    Calculate multilateral index numbers in PySpark.

    Currently supported: Time Product Dummy (tpd) and Time Hedonic Dummy (tdh).

    Parameters
    ----------
    df : Spark DataFrame
        Contains price and quantity columns, a time series column, and a product
        ID column as a minimum. A characteristics column should also be present
        for hedonic methods.
    price_col : str, defaults to 'price'
        User-defined name for the price column.
    quantity_col : str, defaults to 'quantity'
        User-defined name for the quantity column.
    date_col : str, defaults to 'date'
        User-defined name for the date column.
    product_id_col: str, defaults to 'id'
        User-defined name for the product id column.
    characteristics: list of str, defaults to None
        The names of the characteristics columns.
    method: str
        Options: {tpd', 'tdh', 'gk'}
        The multilateral method to apply.

    Returns
    -------
    pd.DataFrame
        A pandas dataframe of the index values.
    """
    method = method.lower()

    if method not in {'tpd', 'tdh', 'gk'}:
        raise ValueError(
            "Invalid method or not implemented yet."
        )
    
    # Get timeseries for output index.;;
    time_series = [i.month for i in df.select(date_col).distinct().collect()]
    
    # Calculate weights for each item in each period.
    df = df.withColumn(
        'weights',
        _weights_calc_pyspark(price_col, quantity_col, date_col)
    )

    if method == 'gk':
        index_vals = geary_khamis_pyspark(df, price_col, date_col, product_id_col)
    if method == 'tpd':
        index_vals = time_dummy_pyspark(
            df,
            len(time_series),
            price_col,
            date_col,
            product_id_col,
        )
    elif method == 'tdh':
        if not characteristics:
            raise ValueError(
                "Characteristics required for TDH."
            )
        else: 
            index_vals = time_dummy_pyspark(
                df,
                len(time_series),
                price_col,
                date_col,
                product_id_col,
                characteristics,
            )

    return (
        pd.DataFrame(
            index_vals, 
            index=time_series
        )
        .rename({0: 'index_value'}, axis=1)
    )




from typing import Sequence

from pyspark.ml import Pipeline
from pyspark.ml.regression import (
    LinearRegression,
    LinearRegressionModel,
)
from pyspark.ml.feature import (
    StringIndexer,
    OneHotEncoder,
    VectorAssembler,
)
from pyspark.sql import DataFrame as SparkDF

__author__ = ['Dr. Usman Kayani']

def wls_pyspark(
    df: SparkDF,
    dependent_var: str,
    independent_vars: Sequence[str],
) -> LinearRegressionModel:
    """Perform weighted least squares regression in PySpark.

    Parameters
    ----------
    df : SparkDF
        Contains columns for each object in the formula and a weights column as
        a minimum.
    dependent_var: str
        The dependent variable for the regression.
    independent_vars: list of str
        The independent variables for the regression.

    Returns
    -------
    LinearRegressionModel
        Weighted linear regression model derived from a least-squares fit.

    """
    # Set up stages for pipeline.
    indexers, encoders, vec_cols = [], [], []
    for column in independent_vars:
        # Map the string cols of labels to numeric values.
        indexers.append(
            StringIndexer(
                inputCol=column,
                outputCol=f'{column}_numeric',
                stringOrderType='alphabetDesc',
            )
            .fit(df)
        )

        # Encode the numeric values to dummy variables.
        encoders.append(
            OneHotEncoder(
                inputCol=f'{column}_numeric',
                outputCol=f'{column}_vector',
            )
        )
        vec_cols.append(f'{column}_vector')

    # Transform vectors for dummy vars into a single vector (features).
    assembler = VectorAssembler(inputCols=vec_cols, outputCol="features")
    # Create a pipeline to perform the sequence of stages above.
    pipeline = Pipeline(stages=indexers + encoders + [assembler])

    # Fit & transform the dataframe according to pipeline for model.
    model_df = pipeline.fit(df).transform(df)

    # WLS regression with labels, features & weights -> fit model.
    wls_model = LinearRegression(
        labelCol=dependent_var,
        featuresCol='features',
        weightCol='weights',
    )
    return wls_model.fit(model_df)


"""
Provides the following multilateral methods:

* :func:`time_dummy_pyspark`
* :func:`geary_khamis_pyspark`

The TDH/TPD methods are model-based multilateral index number methods
which have been proposed to incorporate scanner data. They are part of
many multilateral methods motivated by an attempt to minimize the risk
of chain drift, particularly within a window, while maximizing the
number of matches in the data.

TDH index is used when information on item characteristics are
available, and the TPD index when this information is lacking. The
TDH produces an explicit hedonic price index, while the TPD produces
an implicit hedonic price index, which are both estimated on the
pooled data of one or more periods via an application of expenditure
shares weighted least squares regression.
"""

from typing import List, Sequence, Optional

import numpy as np
from pyspark.mllib.linalg.distributed import (
    IndexedRow,
    IndexedRowMatrix,
    CoordinateMatrix,
    MatrixEntry,
    DenseMatrix,
)
from pyspark.ml.feature import VectorAssembler
from pyspark.sql import (
    DataFrame as SparkDF,
    functions as F,
)
from pyspark import SparkContext

from .weighted_least_squares import wls_pyspark

__author__ = ['Dr. Usman Kayani']

def time_dummy_pyspark(
    df: SparkDF,
    Number_of_periods: int,
    price_col: str = 'price',
    date_col: str = 'month',
    product_id_col: str = 'id',
    characteristics: Optional[Sequence[str]] = None,
) -> List:
    """Obtain the time dummy indices for a given dataframe in PySpark.

    Calculates the time dummy indices using a formula with weighted least
    squares regression.  When passed with characteristics, this function returns
    the Time Dummy Hedonic indices. When passed without it returns the Time
    Product Dummy indices.
    """  
    # Calculate logarithm of the prices for each item.
    df = df.withColumn('log_price', F.log(price_col))

    non_time_vars = characteristics if characteristics else [product_id_col]

    # WLS regression with labels, features & weights -> fit model.
    model = (
        wls_pyspark(
            df,
            dependent_var='log_price',
            independent_vars=[date_col, *non_time_vars],
        )
    )

    # Extracting time dummy coefficients.
    time_dummy_coeff = model.coefficients[:Number_of_periods-1][::-1]
    
    # Get indices from the time dummy coefficients & set first = 1.
    return [1, *np.exp(time_dummy_coeff)]

def geary_khamis_pyspark(
    df: SparkDF,
    sc: SparkContext,
    price_col: str = 'price',
    quantity_col: str = 'quantity',
    date_col: str = 'month',
    product_id_col: str = 'id',
) -> List:
    """Obtain the Geary-Khamis indices for a given dataframe in PySpark.

    Calculates the Geary-Khamis indices using matrix operations.
    """
    pivoted_df = (
        df.groupby(product_id_col)
        .pivot(date_col)
    )

    matrix_dfs = []

    for col in [price_col, quantity_col, 'weights']:
        matrix_dfs.append(
            pivoted_df
            .avg(col)
            .sort(product_id_col)
            .drop(product_id_col)
            .fillna(0)
        )
    cols = matrix_dfs[0].columns
    
    vectors = []
    for mdf in matrix_dfs:
        vectors.append(
            VectorAssembler(inputCols=cols, outputCol='vec')
            .transform(mdf)
            .select('vec')
            .collect()
        )
    
    N = len(vectors[0])
    M = len(cols)

    matrices = []
    for vec in vectors:
        matrices.append(
            IndexedRowMatrix(
                sc
                .range(N)
                .map(lambda i: IndexedRow(i, vec[i][0].array))
            )
        )
    
    qsum_arr = matrices[1].rows.map(lambda row: row.vector.sum()).collect()
    qsum_inv_mat = CoordinateMatrix(
        sc
        .range(N)
        .map(lambda i: MatrixEntry(i, i, 1/qsum_arr[i])), N, N
    )

    C = (
        qsum_inv_mat.toBlockMatrix()
        .multiply(matrices[2].toBlockMatrix())
        .multiply(matrices[1].toBlockMatrix().transpose())
    )


    Identity = (
        CoordinateMatrix(
            sc
            .range(N)
            .map(lambda i: MatrixEntry(i, i, 1.0)), N, N
        )
        .toBlockMatrix()
    )

    R = (
        CoordinateMatrix(
            sc
            .range(N)
            .map(lambda i: MatrixEntry(0, i, 1.0)), N, N
        )
        .toBlockMatrix()
    )

    svd = (
        R.add(Identity).subtract(C)
        .toIndexedRowMatrix()
        .computeSVD(N, True)
    )

    s = svd.s
    V = svd.V
    U = svd.U

    V_ = DenseMatrix(N, N, V.toArray().transpose().reshape(-1))
    invS = DenseMatrix(len(s), len(s), np.diag(1/s).reshape(-1))

    c_ = (
        CoordinateMatrix(
            sc
            .range(N)
            .map(lambda i: MatrixEntry(0, 0, 1/N)), N, 1
        )
        .toBlockMatrix()
        .transpose()
    )

    b = (
        c_.multiply(U.multiply(invS).multiply(V_).toBlockMatrix())
        .toLocalMatrix()
        .toArray()
    )

    mat_prices_ = (
        matrices[0]
        .toBlockMatrix()
        .transpose()
        .toLocalMatrix()
        .toArray()
    )

    mat_quantities_ = (
        matrices[1]
        .toBlockMatrix()
        .transpose()
        .toLocalMatrix()
        .toArray()
    )

    price_levels = []
    for i in range(M):
        val = mat_prices_[i].dot(mat_quantities_[i])/mat_quantities_[i].dot(b[0])
        price_levels.append(val)

    initial_pl = price_levels[0]

    return [pl/initial_pl for pl in price_levels]